@@ -1,5 +1,11 @@
 /*
- * Copyright (C) 2003 _INRIA_
+ * Copyright (C) 2003 INRIA
+ *
+ *	INRIA
+ *	Domaine de Voluceau
+ *	Rocquencourt - B.P. 105
+ *	78153 Le Chesnay Cedex - France
+ *
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +35,7 @@
 #include <signal.h>
 #include <memory.h>
 #include <string.h>
+#include <errno.h>
 
 #include <rfid.h>
 #include <cgic.h>
@@ -56,67 +63,84 @@ static void report_error(rfid_t* rfid, const char* message)
 
 /*
  */
-static int callback(rfid_t* rfid, const rfid_transponder_t* transponder, void* data)
+
+static int transponder_callback(rfid_t* rfid, int event, const rfid_transponder_t* transponder, void* data)
 {
   rfid_cgi_t* rfid_cgi = (rfid_cgi_t*)data;
-  const s6350_transponder_t* info = &transponder->info;
+  char* id = 0;
+  char* event_string = 0;
 
   if(rfid_cgi->max_count > 0)
     rfid_cgi->max_count--;
 
   if(rfid_cgi->verbose) {
-    fprintf(cgiOut, "\
-# ========================================\n\
-# transponder_id = 0x%02x%02x%02x%02x\n\
-# manufacturer = 0x%02x\n\
-# version_number = 0x%04x\n\
-# blocks = %04d\n\
-# bytes_per_block = %d\n\
-",
-	    info->transponder_id[3],
-	    info->transponder_id[2],
-	    info->transponder_id[1],
-	    info->transponder_id[0],
-	    info->manufacturer,
-	    info->version_number,
-	    info->blocks,
-	    info->bytes_per_block
-	    );
-
-    {
-      int i;
-      fprintf(cgiOut, "# ");
-      for(i = 0; i < transponder->data_length; i++)
-	fprintf(cgiOut, "%c(0x%02x)/", transponder->data[i], transponder->data[i]);
-      fprintf(cgiOut, "\n");
-    }
+    int i;
+    fprintf(cgiOut, "# ");
+    for(i = 0; i < transponder->data_length; i++)
+      fprintf(cgiOut, "%c(0x%02x)/", transponder->data[i], transponder->data[i]);
+    fprintf(cgiOut, "\n");
   }
 
-  fprintf(cgiOut, "%02x%02x%02x%02x.%02x.%04x\t%04d\t%s\n",
-	  info->transponder_id[3],
-	  info->transponder_id[2],
-	  info->transponder_id[1],
-	  info->transponder_id[0],
-	  info->manufacturer,
-	  info->version_number,
-	  info->blocks * info->bytes_per_block,
+  switch(event) {
+  case RFID_EVENT_NEW:
+    event_string = "N";
+    break;
+  case RFID_EVENT_DELETE:
+    event_string = "D";
+    break;
+  case RFID_EVENT_AGAIN:
+    event_string = "A";
+    break;
+  default:
+    event_string = "(unknown event)";
+    break;
+  }
+  
+  rfid_transponder_id_get(rfid, transponder, &id);
+
+  fprintf(cgiOut, "%s\t%s\t%04d\t%s\n",
+	  event_string,
+	  id,
+	  transponder->blocks * transponder->bytes_per_block,
 	  transponder->data);
 
   fflush(cgiOut);
 
+  free(id);
+
   if(rfid_cgi->max_count >= 0)
     return rfid_cgi->max_count == 0 ? RFID_CALLBACK_END : RFID_CALLBACK_OK;
   else
     return RFID_CALLBACK_OK;
 }
 
+static int inventory_callback(rfid_t* rfid, rfid_transponder_t** transponders, int transponders_length, void* data)
+{
+  int i;
+
+  fprintf(cgiOut, "I");
+
+  for(i = 0; i < transponders_length; i++) {
+    char* id = 0;
+    rfid_transponder_id_get(rfid, transponders[i], &id);
+    fprintf(cgiOut, "\t%s", id);
+    free(id);
+  }
+
+  fprintf(cgiOut, "\n");
+
+  fflush(cgiOut);
+
+  return RFID_CALLBACK_OK;
+}
+
 /*
  * Implements read=1, infinite loop reading commands from RFID transponders.
  */
 static int read_daemon(rfid_cgi_t* rfid_cgi)
 {
-  if(rfid_read(rfid_cgi->rfid, callback, rfid_cgi) < 0) {
-    report_error(rfid_cgi->rfid, "rfid_read");
+  if(rfid_event_loop(rfid_cgi->rfid, RFID_EVENT_NEW|RFID_EVENT_DELETE|RFID_EVENT_INVENTORY, transponder_callback, inventory_callback, rfid_cgi) < 0) {
+    report_error(rfid_cgi->rfid, "rfid_event_loop");
     return -1;
   }
 
@@ -124,28 +148,33 @@ static int read_daemon(rfid_cgi_t* rfid_cgi)
 }
 
 /*
- * Wait for a RFID transponder to be presented to the RFID reader
- * and store the arguments (white space separated) into it.
  */
-static int write_command(rfid_cgi_t* rfid_cgi, char* data)
+static int write_string(rfid_cgi_t* rfid_cgi, char* id, char* data)
 {
   rfid_t* rfid = rfid_cgi->rfid;
-  rfid_transponder_t transponder;
+  rfid_transponder_t* transponder = rfid_transponder_alloc(rfid);
 
-  memset(transponder.data, '\0', RFID_TRANSPONDER_DATA_SIZE);
-  memcpy(transponder.data, data, RFID_TRANSPONDER_DATA_SIZE);
-  transponder.data_length = RFID_TRANSPONDER_DATA_SIZE;
+  strncpy(transponder->data, data, RFID_TRANSPONDER_DATA_SIZE);
+  transponder->data_length = RFID_TRANSPONDER_DATA_SIZE;
+  if(strlen(id) > 0) {
+    if(rfid_transponder_id_set(rfid, transponder, id) < 0) {
+      report_error(rfid, "rfid_transponder_id_set");
+      return -1;
+    }
+  }
 
   if(rfid_cgi->verbose) {
-    fprintf(cgiOut, "# Attempting to write '%s' to a transponder ... ", transponder.data);
+    fprintf(cgiOut, "# Attempting to write '%s' to a %s transponder ... ", transponder->data, (strlen(id) > 0 ? id : "any"));
     fflush(cgiOut);
   }
   
-  if(rfid_write(rfid, &transponder) < 0) {
+  if(rfid_write(rfid, transponder) < 0) {
     report_error(rfid, "rfid_write");
     return -1;
   }
 
+  rfid_transponder_free(rfid, transponder);
+
   if(rfid_cgi->verbose) {
     fprintf(cgiOut, "done\n");
     fflush(cgiOut);
@@ -162,12 +191,14 @@ static void usage(const char* message)
   char* usage = "\
 # usage:\n\
 #\n\
-# rfid-cgi?[help=<n>&][device=<special>&][verbose=1&][read=<n>]\n\
-# rfid-cgi?[help=<n>&][device=<special>&][verbose=1&][write=<data>]\n\
+# rfid-cgi?[help=<n>&][device=<special>&][iso=1&][verbose=1&][read=<n>]\n\
+# rfid-cgi?[help=<n>&][device=<special>&][iso=1&[verbose=1&][write=<data>]\n\
 #    read=<n>           Wait for a transponder and display its content.\n\
 #                       Loop <n> times.\n\
 #    write=<data>       Wait for a transponder and store the data\n\
 #                       given in argument in its memory.\n\
+#    id=<hex>           II.II... RFID transponder identifier.\n\
+#    iso=1              Expect ISO-15693 RFID transponders.\n\
 #    device=<special>   Dialog with RFID reader connected to special\n\
 #                       file <special>. Use %2F instead of /\n\
 #                       (default %2Fdev%2FttyS0).\n\
@@ -177,18 +208,18 @@ static void usage(const char* message)
 #    When read=<n> is set, each tag is reported using the following format,\n\
 #    one per line. Newlines are forbiden and will break the format.\n\
 #    \n\
-#    IIIIIIII.MM.VVVV<tab>BBBB<tab>CCCCCCC...\n\
+#    [DN] <tab>II.II...<tab>BBBB<tab>CCCCCCC...\n\
 #    \n\
 #    Where:\n\
 #    \n\
-#    IIIIIIII  is a 8 digit transponder identification in hexadecimal.\n\
-#    MM        is a 2 digit manufacturer identification in hexadecimal.\n\
-#    VVVV      is a 4 digit version number in hexadecimal.\n\
-#    BBBB      is a is the size of the memory of the transponder as\n\
-#              4 digit decimal number\n\
-#    CCCCC.... is the content of the memory of the transponder in characters\n\
-#              until a newline is found.\n\
-#    <tab>     is a tabulation used as a separator.\n\
+#    [DN]         is one letter describing the event: \n\
+#                 N means the RFID transponder entered the field of the reader,\n\
+#    II.II...     is a transponder identification in dotted hexadecimal.\n\
+#    BBBB         is a is the size of the memory of the transponder as\n\
+#                 4 digit decimal number\n\
+#    CCCCC....    is the content of the memory of the transponder in characters\n\
+#                 until a newline is found.\n\
+#    <tab>        is a tabulation used as a separator.\n\
 #\n\
 ";
   fprintf(stderr, "usage: %s\n", message);
@@ -201,10 +232,14 @@ int cgiMain() {
   int retval = 0;
   int option_read = 0;
   int option_write = 0;
+  int option_iso = 0;
   char data[RFID_TRANSPONDER_DATA_SIZE + 1];
+  char id[RFID_TRANSPONDER_HEX_ID_SIZE + 1];
   rfid_t rfid;
   rfid_cgi_t rfid_cgi;
 
+  memset(id, '\0', RFID_TRANSPONDER_HEX_ID_SIZE + 1);
+  memset(data, '\0', RFID_TRANSPONDER_DATA_SIZE + 1);
   memset(&rfid, '\0', sizeof(rfid_t));
   rfid.device = (char*)malloc(DEVICE_SIZE+1);
 
@@ -224,21 +259,45 @@ int cgiMain() {
   }
   
   if(cgiFormString("device", rfid.device, DEVICE_SIZE) != cgiFormSuccess)
-      strcpy(rfid.device, "/dev/ttyS0");
+    strcpy(rfid.device, "/dev/ttyS0");
+
+  if(cgiFormString("id", id, RFID_TRANSPONDER_HEX_ID_SIZE) != cgiFormSuccess)
+    strcpy(id, "");
 
   if(cgiFormInteger("read", &rfid_cgi.max_count, 1) == cgiFormSuccess) {
     option_read++;
   }
 
-  if(cgiFormString("write", data, RFID_TRANSPONDER_DATA_SIZE) == cgiFormSuccess) {
+  cgiFormInteger("iso", &option_iso, 0);
+
+  switch(cgiFormString("write", data, RFID_TRANSPONDER_DATA_SIZE)) {
+  case cgiFormSuccess:
     option_write++;
+    break;
+
+  case cgiFormTruncated:
+    fprintf(cgiOut, "# Absolute maximum write size is %d\n", RFID_TRANSPONDER_DATA_SIZE);
+    usage("Absolute maximum write size overflow");
+    return -1;
+    break;
+
+  default:
+    break;
   }
 
   if(option_read && option_write) {
     usage("Options read=<n> and write=<data> are mutually exclusive");
     return -1;
   }
-  
+
+  if(strlen(id) > 0 && !option_write) {
+    usage("Options id=<hex> requires write=<data>");
+    return -1;
+  }
+
+  if(option_iso)
+    rfid_options(&rfid, RFID_OPTION_ISO15693);
+
   if(rfid_init(&rfid) < 0) {
     report_error(&rfid, "rfid_init");
     return -1;
@@ -247,9 +306,13 @@ int cgiMain() {
   if(option_read)
     retval = read_daemon(&rfid_cgi);
   else if(option_write) {
-    retval = write_command(&rfid_cgi, data);
-  } else
-    usage("You must specify either write=<data> or read=1");
+    retval = write_string(&rfid_cgi, id, data);
+  } else {
+    usage("You must specify either write=<data> or read=<n>");
+    retval = -1;
+  }
+
+  free(rfid.device);
 
   rfid_end(&rfid);
 
